//-----------------------//
//    COLLISION MERGER   //
//    Mikelan98, 2024    //
//-----------------------//

#include <stdio.h>
#include <string.h>
#include <malloc.h>

/* CONSTANTS */
// Game version definitions
#define DPPT	1
#define HGSS	2

// Target game version
// EDIT ME IF HGSS MAPS ARE TO BE CONVERTED
#define GAME_VERSION	DPPT

// Map tile size
#define MAP_SIZE_X		32
#define MAP_SIZE_Y		32
#define MAP_SIZE		MAP_SIZE_X * MAP_SIZE_Y

// Collision type count
#define COLLISION_TYPE_COUNT		256
#define SOUND_COLLISION_COUNT		16

enum {
	COLLISION_WALKABLE,
	COLLISION_HITTABLE,
	COLLISION_HIT_TYPE_COUNT,
};

#define COLLISION_ID_EMPTY			0x00
#define COLLISION_ID_HIT			0x01

/* STRUCTS */
// Map header (part sizes) format
#if GAME_VERSION == DPPT
typedef struct {
	unsigned __int32 collisionSize;
	unsigned __int32 buildingSize;
	unsigned __int32 modelSize;
	unsigned __int32 terrainSize;
} MAP_HEADER_STRUCT;
#elif GAME_VERSION == HGSS
typedef struct {
	unsigned __int32 collisionSize;
	unsigned __int32 buildingSize;
	unsigned __int32 modelSize;
	unsigned __int32 terrainSize;
	unsigned __int16 soundMagic;
	unsigned __int16 soundSize;
} MAP_HEADER_STRUCT;
#else
#error
#endif

// Map file format
typedef struct {
	MAP_HEADER_STRUCT header;
	unsigned __int8 data[];
} MAP_STRUCT;

/* STATIC */
// Output statistical info data
static unsigned int hitData[COLLISION_TYPE_COUNT][COLLISION_HIT_TYPE_COUNT];
#if GAME_VERSION == HGSS
static unsigned int soundData[COLLISION_TYPE_COUNT][SOUND_COLLISION_COUNT];
#endif

// Default hit types
#if GAME_VERSION == DPPT
#include "DefaultHitTypesDPPT.h"
#elif GAME_VERSION == HGSS
#include "DefaultHitTypesHGSS.h"
#endif

/* PROGRAM */
int main(int argc, char* argv[])
{
	// Program usage print
	if (argc < 2) {
		printf("Collision Merger by Mikelan98\nUsage: CollisionMerger.exe map.bin [...]\nAll the game's map files should be converted at once\n");
		return 0;
	}

	// Create statistical output file
	FILE* file_output = fopen("output.txt", "w");

	for (int i = 1; i < argc; i++) {

		// Read and parse every file passed as argument
		char* path = argv[i];

		FILE* file = fopen(path, "rb");
		if (file == NULL) return -1;

		// Get the current map file size
		fseek(file, 0L, SEEK_END);
		int map_data_size = ftell(file);

		// Allocate memory block for the whole map file
		MAP_STRUCT* map_data = malloc(map_data_size);
		if (map_data == NULL) return -1;

		// Read the map data into the allocated memory block
		fseek(file, 0L, SEEK_SET);
		fread(map_data, map_data_size, 1, file);

		fclose(file);

		// Depending on the game version, the collision data pointer will be get diferently
		// Both will be __int16 pointer since each tile collision takes 2 bytes
#if GAME_VERSION == HGSS
		unsigned __int16* collision_data_p = (__int16*)(&map_data->data[0] + map_data->header.soundSize);
#else
		unsigned __int16* collision_data_p = (__int16*)(&map_data->data[0]);
#endif

		for (int j = 0; j < MAP_SIZE; j++) {
			unsigned __int8 collision_type = collision_data_p[j] & 0xFF;
			int collision_hit = (collision_data_p[j] >> 15) & 1;
#if GAME_VERSION == HGSS
			int collision_sound = (collision_data_p[j] >> 8) & 0x7F;
			if (collision_sound >= SOUND_COLLISION_COUNT) return -1;
#endif
			hitData[collision_type][collision_hit]++;
#if GAME_VERSION == HGSS
			soundData[collision_type][collision_sound]++;
#endif
			// Whenever an abnormal combination of type/hit in a tile collision is found, print a warning line
			// Some maps have such abnormalities in the original games, so these warnings can be ignored when converting original maps
			if (collision_type != COLLISION_ID_EMPTY && DefaultHitTypes[collision_type] != collision_hit) {
				printf("Abnormal collision hit found: Map %s Collision %02X Hit %d\n", path, collision_type, collision_hit);
			}

			if (collision_hit && DefaultHitTypes[collision_type] != collision_hit) {
				/* If a tile collision is marked as hit but the collision type is not inherently hittable, set collision type 01 in the converted map
				This can happen in two scenarios:
				1) Collision type is 00 and is marked as hit (this is a normal hit tile), so it will be normally replaced by the new hit type
				2) An abnormal collision (marked as hit) is found, so it is ignored and replaced by the new hit type */
				collision_data_p[j] = COLLISION_ID_HIT;
			} else {
				/* In any other case, write the current collision type. This can happen if either:
				1) Collision type is inherently not hittable and is correctly marked as no hit
				2) Collision type is inherently hittable, either if it is correctly marked as hit or not
				*/
				collision_data_p[j] = collision_type;
			}
#if GAME_VERSION == HGSS
			collision_data_p[j] |= collision_sound << 8;
#endif
		}

		// Write new, converted map file
		fopen_s(&file, path, "wb");
		if (file == NULL) return -1;
		fwrite(map_data, map_data_size, 1, file);
		fclose(file);

		free(map_data);
	}

	// Print hit statistical data
	fprintf(file_output, "Type\tNo Hit\tHit\n");
	for (int type = 0; type < COLLISION_TYPE_COUNT; type++) {
		fprintf(file_output, "%02X\t%d\t%d\n", type, hitData[type][COLLISION_WALKABLE], hitData[type][COLLISION_HITTABLE]);
	}

	// For HGSS only, print sound statistical data too
#if GAME_VERSION == HGSS
	fprintf(file_output, "\nType");
	for (int i = 0; i < SOUND_COLLISION_COUNT; i++) {
		fprintf(file_output, "\tSound%d", i);
	}
	fprintf(file_output, "\n");
	for (int type = 0; type < COLLISION_TYPE_COUNT; type++) {
		fprintf(file_output, "%02X", type);
		for (int sound = 0; sound < SOUND_COLLISION_COUNT; sound++) {
			fprintf(file_output, "\t%d", soundData[type][sound]);
		}
		fprintf(file_output, "\n");
	}
#endif

	fclose(file_output);

	printf("Succesfully converted %d maps collisions\n", argc - 1);
	system("pause");
	return 0;
}
